"""
Elliptic curves
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import sage.rings.arith as arith
import sage.rings.rings as rings
import sage.misc.errors as errors
import sage.misc.misc as misc
import sage.modular.modform as modform
import sage.rings.rings as rings
import sage.tables.elliptic_curves 
import sage.pari.py_pari as py_pari
pari = py_pari.pari
import sage.functions.transcendental as transcendental
import math
factor = arith.factor
sqrt = math.sqrt
exp = math.exp
mul = misc.mul
next_prime = arith.next_prime

try:
    import mwrank.mwrank as mwrank
    MWRANK = True
except ImportError:
    MWRANK = False
    pass

oo = rings.infinity       # infinity
Q = rings.RationalField()         # rational numbers
O = rings.O         # big oh
Z = rings.IntegerRing()
C = rings.ComplexField()
R = rings.RealField()

class EllipticCurve(object):
    def __init__(self, ainvs):
        self.__np = {}
        if isinstance(ainvs, str):
            X = sage.tables.elliptic_curves.Database().curve(ainvs)
            self.__ainvs = [Q(a) for a in X.a_invariants]
            self.__base_field = Q
            self.__analytic_rank = X.rank
            self.__rank = X.rank
            self.__conductor = Z(X.conductor)
            self.__torsion_order = X.torsion
            self.__cremona_label = "%s%s%s"%(X.conductor,X.iso_class,X.number)
            self.__tamagawa_number = X.tamagawa
            return
        elif isinstance(ainvs, sage.tables.elliptic_curves.EllipticCurve):
            X = ainvs
            self.__ainvs = X.a_invariants
            self.__base_field = Q
            self.__analytic_rank = X.rank
            self.__rank = X.rank
            self.__conductor = Z(X.conductor)
            self.__torsion_order = X.torsion
            self.__cremona_label = "%s%s%s"%(X.conductor,X.iso_class,X.number)
            self.__tamagawa_number = X.tamagawa
            return
        if len(ainvs) != 2 and len(ainvs) != 5:
            raise TypeError, "Number of invariants (=%s) must be 2 or 5."%len(ainvs)
        if not isinstance(ainvs[0], rings.Element) or not ainvs[0].parent().is_field():
            try:
                ainvs = [Q(a) for a in ainvs]
            except TypeError:
                raise TypeError, "Defining coefficients (=%s) must lie in a field."%ainvs
        self.__base_field = ainvs[0].parent()
        if len(ainvs) == 2:
            ainvs = [0,0,0] + ainvs
        self.__ainvs = [self.__base_field(x) for x in ainvs]

    def __repr__(self):
        a = self.__ainvs
        #return "y^2 + %s*x*y + %s*y = x^3 + %s*x^2 + %s*x + %s"%\
        #       (a[0], a[2], a[1], a[3], a[4])
        s = "Elliptic Curve defined by "
        s += "y^2 "
        if a[0] == -1:
            s += "- x*y "
        elif a[0] == 1:
            s += "+ x*y "
        elif a[0] != 0:
            s += "+ %s*x*y "%a[0]
        if a[2] == -1:
            s += " - y"
        elif a[2] == 1:
            s += "+ y"
        elif a[2] != 0:
            s += "+ %s*y"%a[2]
        s += " = x^3 "
        if a[1] == -1:
            s += "- x^2 "
        elif a[1] == 1:
            s += "+ x^2 "
        elif a[1] != 0:
            s += "+ %s*x^2 "%a[1]
        if a[3] == -1:
            s += "- x "
        elif a[3] == 1:
            s += "+ x "
        elif a[3] != 0:
            s += "+ %s*x "%a[3]
        if a[4] == -1:
            s += "-1"
        elif a[4] == 1:
            s += "+1"
        elif a[4] != 0:
            s += "+ %s"%a[4]
        s = s.replace("+ -","- ")
        s += " over %s"%self.base_field()
        return s

    def __contains__(self, P):
        if not isinstance(P, Point):
            P = self(P)
        if P.curve() == self:
            return True
        x, y, a = P[0], P[1], self.__ainvs
        return y**2 + a[0]*x*y + a[2]*y == x**3 + a[1]*x**2 + a[3]*x + a[4]

    def __call__(self, x=oo, y=oo):
        if isinstance(x,Point):
            if x.curve() == self:
                return x
            else:
                x, y = x[0], x[1]
        elif isinstance(x,list):
            if len(x) != 2:
                raise errors.UsageError("List must have length 2.")
            x, y = x[0], x[1]
        return Point(self, x, y)

    def __getitem__(self, n):
        """
        """
        raise NotImplementedError, "not implemented."

    def ainvs(self):
        """
        The a-invariants of the elliptic curve.
        """
        return self.__ainvs

    def a_invariants(self):
        """
        The a-invariants of the elliptic curve.
        """
        return self.__ainvs

    def base_field(self):
        return self.__base_field

    def a1(self):
        return self.__ainvs[0]
    
    def a2(self):
        return self.__ainvs[1]
    
    def a3(self):
        return self.__ainvs[2]
    
    def a4(self):
        return self.__ainvs[3]
    
    def a6(self):
        return self.__ainvs[4]

    

    def Np(self, p):
        """
        The number of points on E modulo p, where p is a prime
        of good reduction.
        """
        return p+1-self.ap(p)



    ####################################################################
    #  Access to PARI curves related to this curve.
    ####################################################################
        
    def pari_curve(self):
        """
        Return the PARI curve corresponding to this elliptic curve.
        """
        try:
            return self.__pari_curve
        except AttributeError:
            if not isinstance(self.base_field(), rings.RationalField):
                raise NotImplementedError, "This computation is only implemented for elliptic curves over the rational number field."
            self.__pari_curve = py_pari.pari(self.__ainvs).ellinit()
        return self.__pari_curve

    def pari_mincurve(self):
        """
        Return the PARI curve corresponding to a minimal model
        for this elliptic curve.
        """
        try:
            return self.__pari_mincurve
        except AttributeError:
            if not isinstance(self.base_field(), rings.RationalField):
                raise NotImplementedError, "This computation is only implemented for elliptic curves over the rational number field."
            e = self.pari_curve()
            R = e.ellglobalred()
            self.__conductor = Z(R[0])
            self.__pari_mincurve = e.ellchangecurve(R[1])
        return self.__pari_mincurve
    
    ####################################################################
    #  Access to mwrank
    ####################################################################
    def mwrank_curve(self):
        if not isinstance(self.__base_field, rings.RationalField):
            raise ArithmeticError, "mwrank is only for curves over the rational numbers."
        try:
            return self.__mwrank_curve
        except AttributeError:
            if MWRANK:
                self.__mwrank_curve = mwrank.EllipticCurve(self.__ainvs)
            else:
                raise ImportError, "the mwrank package must be installed for requested functionality."
        return self.__mwrank_curve
    

    ####################################################################
    #  Etc.
    ####################################################################

    def anlist(self, n):
        """
        The Fourier coefficients up to and including a_n of the
        modular form attached to this elliptic curve.  The ith element
        of the return list is a[i].
        

        EXAMPLES:
            >>> E = EllipticCurve([0, -1, 1, -10, -20])
            >>> E.anlist(3)
            [0, 1, -2, -1]
            
            >>> E = EllipticCurve([0,1])
            >>> E.anlist(20)
            [0, 1, 0, 0, 0, 0, 0, -4, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 8, 0]
        """
        n = int(n)
        try:
            if len(self.__anlist) > n:
                return self.__anlist[:n+1]
        except AttributeError:
            pass
        E = self.pari_mincurve()
        
        # There is some overheard associated with coercing the PARI
        # list back to Python, but it's not bad.  It's better to do it
        # this way instead of trying to eval the whole list, since the
        # int conversion is done very sensibly.  NOTE: This would fail
        # if a_n won't fit in a C int, i.e., is bigger than
        # 2147483648; however, we wouldn't realistically compute
        # anlist for n that large anyways.
        #
        # Some relevant timings:
        #
        # E <--> [0, 1, 1, -2, 0]   389A
        #  E = EllipticCurve([0, 1, 1, -2, 0]);   // SAGE or MAGMA
        #  e = E.pari_mincurve()
        #  f = ellinit([0,1,1,-2,0]);
        #
        #  Computation                                              Time (1.6Ghz Pentium-4m laptop)
        #  time v:=TracesOfFrobenius(E,10000);  // MAGMA            0.120
        #  gettime;v=ellan(f,10000);gettime/1000                    0.046
        #  time v=e.ellan (10000)                                   0.04
        #  time v=E.anlist(10000)                                   0.07
        
        #  time v:=TracesOfFrobenius(E,100000);  // MAGMA           1.620
        #  gettime;v=ellan(f,100000);gettime/1000                   0.676
        #  time v=e.ellan (100000)                                  0.7
        #  time v=E.anlist(100000)                                  0.83

        #  time v:=TracesOfFrobenius(E,1000000);  // MAGMA          20.850
        #  gettime;v=ellan(f,1000000);gettime/1000                  9.238
        #  time v=e.ellan (1000000)                                 9.61
        #  time v=E.anlist(1000000)                                 10.95  (13.171 in cygwin vmware)
        
        #  time v:=TracesOfFrobenius(E,10000000);  //MAGMA          257.850
        #  gettime;v=ellan(f,10000000);gettime/1000      FAILS no matter how many allocatemem()'s!!
        #  time v=e.ellan (10000000)                                139.37  
        #  time v=E.anlist(10000000)                                136.32
        #
        #  The last SAGE comp retries with stack size 40MB,
        #  80MB, 160MB, and succeeds last time.  It's very interesting that this
        #  last computation is *not* possible in GP, but works in py_pari!
        #
        if n >= 2147483648:
            raise RuntimeError, "anlist: n must be < 2147483648."
        
        self.__anlist = [0] + [int(x) for x in E.ellan(n)]
        return self.__anlist

    def analytic_rank(self):
        try:
            return self.__analytic_rank
        except AttributeError:
            raise NotImplementedError
            #x = pari.ellanalyticrank(self)
            #self.__analytic_rank = x[0]
            return self.__analytic_rank

    def rank(self):
        try:
            return self.__rank
        except AttributeError:
            C = self.mwrank_curve()
            r = C.rank()
            if not C.certain():
                raise RuntimeError, "Unable to compute the rank with certainty (lower bound=%s)."%r
            self.__rank = r
        return self.__rank

    def gens(self):
        try:
            return self.__gens
        except AttributeError:
            C = self.mwrank_curve()
            G = C.gens()
            if not C.certain():
                raise RuntimeError, "Unable to compute the rank, hence generators, with certainty (lower bound=%s)."%C.rank()
            self.__gens = [Point(self, x, check=False) for x in G]
        return self.__gens
        
    def regulator(self):
        try:
            return self.__regulator
        except AttributeError:
            C = self.mwrank_curve()
            R = C.regulator()
            if not C.certain():
                raise RuntimeError, "Unable to compute the rank, hence regulator, with certainty (lower bound=%s)."%C.rank()
            self.__regulator = R
        return self.__regulator

    def two_selmer_rank(self):
        try:
            return self.__two_selmer_rank
        except AttributeError:
            C = self.mwrank_curve()
            r = C.selmer_rank()
            self.__two_selmer_rank = r
            return self.__two_selmer_rank
    
            
    def an(self, n):
        """
        The n-th Fourier coefficient of the modular form corresponding
        to this elliptic curve, where n is a positive integer.
        """
        return int(self.pari_mincurve().ellak(n))

    def ap(self, p):
        """
        The p-th Fourier coefficient of the modular form corresponding
        to this elliptic curve, where p is prime.
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        return int(self.pari_mincurve().ellap(p))

    def conductor(self):
        """
        Returns the conductor of the elliptic curve.
        """
        try:
            return self.__conductor
        except AttributeError:
            self.pari_mincurve() # computes conductor as side effect
        return self.__conductor

    def quadratic_twist(self, D):
        a = self.__ainvs
        return EllipticCurve([0, 0, 0, \
                              -27*D**2*a[0]**4 - 216*D**2*a[0]**2*a[1] + 648*D**2*a[0]*a[2] - 432*D**2*a[1]**2 + 1296*D**2*a[3], \
                              54*D**3*a[0]**6 + 648*D**3*a[0]**4*a[1] - 1944*D**3*a[0]**3*a[2] + 2592*D**3*a[0]**2*a[1]**2 - 3888*D**3*a[0]**2*a[3] - 7776*D**3*a[0]*a[1]*a[2] + 3456*D**3*a[1]**3 - 15552*D**3*a[1]*a[3] + 11664*D**3*a[2]**2 + 46656*D**3*a[4]]).minimal_model()

    def minimal_model(self):
        F = self.pari_mincurve()
        G = [Q(F[i]) for i in range(5)]
        return EllipticCurve(G)

    def kodaira_type(self, p):
        """
        Local Kodaira type of the elliptic curve at p.

        1 means good reduction (type I_0), 2, 3 and 4 mean types II, III and IV, respectively,
        4 + nu with nu > 0 means type I_{nu}; finally the opposite values -1, -2, etc. refer
        to the starred types I_0^*, II^*, etc.  
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        try:
            self.__kodaira_type
        except AttributeError:
            self.__kodaira_type = {}
            self.__tamagawa_number = {}
        if not self.__kodaira_type.has_key(p):
            v = self.pari_mincurve().elllocalred(p)
            self.__kodaira_type[p] = str(v[1])
            self.__tamagawa_number[p] = int(v[3])
        return self.__kodaira_type[p]

    def tamagawa_number(self, p):
        """
        The Tamagawa number of the elliptic curve at p.
        """
        if not arith.is_prime(p):
            raise ArithmeticError, "p must be prime"
        try:
            return self.__tamagawa_number[p]
        except (AttributeError, KeyError):
            self.kodaira_type(p)
            return self.__tamagawa_number[p]

    def omega(self):
        """
        Return a basis for the period lattice of the elliptic curve
        over Q as a 2-tuple.

        The basis has the form [omega_1, omega_2], where
        Im(omega_1/omega_2) > 0 and omega_1 is real.
        
        INPUT:
            -- an elliptic curve
        OUTPUT:
            omega_1 -- complex number
            omega_2 -- complex number
        """
        return tuple(self.pari_curve().omega().python())

    def volume(self):
        """
        Return the volume of the period lattice of the elliptic curve.
        """
        w1,w2 = self.omega()
        return (w1*w2.imag()).real()

    def Lseries_at1(self, k=0):
        """
        Compute L(E,1) using k terms of the series for L(E,1) form
        page 406 of Henri Cohen's book "A Course in Computational
        Algebraic Number Theory".  If the k argument is not specified,
        then it defaults to sqrt(N), where N is the conductor.

        The real precision of the computation is the precision of Python
        floats.

        INPUT:
            k -- (optional) an integer, defaults to sqrt(N).
            
        OUTPUT:
            float -- L(E,1)
            float -- a bound on the error in the approximation; this
                     is a proveably correct upper bound on the sum
                     of the tail end of the series used to compute L(E,1).

        EXAMPLES:
        

        ALGORITHM:
            1. Compute the root number eps.  If it is -1, return 0.
            
            2. Compute the Fourier coefficients a_n, for n up to and
               including k.
               
            3. Compute the sum
            
                 2 * sum_{n=1}^{k} (a_n / n) * exp(-2*pi*n/Sqrt(N)),
                 
               where N is the conductor of E.
               
            4. Compute a bound on the tail end of the series, which is
            
                 2 * e^(-2 * pi * (k+1) / sqrt(N)) / (1 - e^(-2*pi/sqrt(N))).
                 
               For a proof see [Grigov-Jorza-Patrascu-Patrikis-Stein].
        """
        if self.root_number() == -1:
            return 0
        sqrtN = sqrt(self.conductor())  
        if k == 0: k = int(math.ceil(sqrtN))
        an = self.anlist(k)           # list of C ints
        # Compute z = e^(-2pi/sqrt(N))
        pi = 3.14159265358979323846
        z = exp(-2*pi/sqrtN)
        zpow = z
        s = 0.0
        for n in xrange(1,k+1):
            s += (zpow * an[n])/n
            zpow *= z

        error = 2*zpow / (1 - z)
        
        return 2*s, error

    def Lseries_deriv_at1(self, k=0):
        """
        Compute L'(E,1) using k terms of the series for L'(E,1) form
        page 406 of Henri Cohen's book"A Course in Computational
        Algebraic Number Theory".

        The real precision of the computation is the precision of
        Python floats.

        INPUT:

        OUTPUT:

        ALGORITHM:
            1. Compute the root number eps.  If it is 1, return 0.

            2. Compute the Fourier coefficients a_n, for n up to and
               including k.
               
            3. Compute the sum
            
                 2 * sum_{n=1}^{k} (a_n / n) * E_1(2*pi*n/Sqrt(N)),
                 
               where N is the conductor of E, and E_1 is the
               exponential integral function.

            4. Compute a bound on the tail end of the series, which is
            
                 2 * e^(-2 * pi * (k+1) / sqrt(N)) / (1 - e^(-2*pi/sqrt(N))).
                 
               For a proof see [Grigov-Jorza-Patrascu-Patrikis-Stein].  This
               is exactly the same as the bound for the approximation to
               L(E,1) produced by Lseries_at1.
        """
        if self.root_number() == 1: return 0
        sqrtN = sqrt(self.conductor())  
        if k == 0: k = int(math.ceil(sqrtN))
        an = self.anlist(k)           # list of C ints
        # Compute z = e^(-2pi/sqrt(N))
        pi = 3.14159265358979323846
        v = transcendental.exponential_integral_1(2*pi/sqrtN, k)
        L = 2*float(sum([ (v[n-1] * an[n])/n for n in xrange(1,k+1)]))
        error = 2*exp(-2*pi*(k+1)/sqrtN)/(1-exp(-2*pi/sqrtN))
        return L, error

    def Lseries(self, s):
        """
        Returns the value of the L-series of the elliptic curve E at s, where s
        must be a real number.  

        NOTES:
        (1) If the conductor of the curve is large, say > 10^12, then this function
        will take a long time, since it uses an O(sqrt(N)) algorithm.
        
        (2) There is an algorithm to compute L(E,s) for any complex s, but it involves
        computing the incomplete Gamma function, which is not implemented in MANIN.

        EXAMPLES:
            >>> E = EllipticCurve([1,2,3,4,5])
            
            >> E.Lseries(1)
            0.E-19

            The following may have low order bits that are different
            on your machine.
            >> E.Lseries(1.1)    
            0.28549100767814812844289234320
        """
        return self.pari_mincurve().elllseries(s).python()

    def Lambda(self, s, prec):
        """
        Returns the value of the Lambda-series of the elliptic curve E
        at s can be any complex number.  Use Lambda(s) if s is real,
        since it will be much faster.

        WARNING: This function uses sage.functions.transcendental.gamma_inc,
        which currently calls mathematica.

        EXAMPLES:
        """
        s = C(s)
        N = self.conductor()
        pi = transcendental.pi()
        Gamma = transcendental.gamma
        Gamma_inc = transcendental.gamma_inc
        a = self.anlist(prec)
        eps = self.root_number()
        sqrtN = arith.sqrt(N)
        def F(n, t):
            return Gamma_inc(t+1, 2*pi*n/sqrtN) * C(sqrtN/(2*pi*n))**(t+1)
        return sum([a[n]*(F(n,s-1) + eps*F(n,1-s)) for n in xrange(1,prec+1)])

    def Lseries_extended(self, s, prec):
        """
        Returns the value of the L-series of the elliptic curve E
        at s can be any complex number using prec terms of the
        power series expansion.

        WARNING: This function uses sage.functions.transcendental.gamma_inc,
        which currently calls mathematica.

        EXAMPLES:

        """
        try:
            s = C(s)
        except TypeError:
            raise TypeError, "Input argument %s must be coercible to a complex number"%s
        prec = int(prec)
        if abs(s.imag()) < R(0.0000000000001):
            return self.Lseries(s.real())
        N = self.conductor()
        pi = transcendental.pi()
        Gamma = transcendental.gamma
        Gamma_inc = transcendental.gamma_inc
        a = self.anlist(prec)
        eps = self.root_number()
        sqrtN = arith.sqrt(N)
        def F(n, t):
            return Gamma_inc(t+1, 2*pi*n/sqrtN) * C(sqrtN/(2*pi*n))**(t+1)
        return C(N)**(-s/2) * C(2*pi)**s * Gamma(s)**(-1)\
               * sum([a[n]*(F(n,s-1) + eps*F(n,1-s)) for n in xrange(1,prec+1)])
               
    def sigma(self, z, flag=0):
        """
        Returns the value of the Weierstrass sigma function of the lattice
        associated to this elliptic curve E.
        INPUT:
            z -- a complex number
            flag -- 0 - default ???
                    1 - computes an arbitrary determination of log(sigma(z))
                    2, 3 - same using the product expansion instead of theta series.
                           ???
        OUTPUT:
            a complex number

        NOTE: The reason for the ???'s above, is that the PARI documentation for
              ellsigma is very vague.
        """
        return self.pari_curve().ellsigma(z, flag)

    def modular_parametrization(self):
        """
        Computes and returns ...
        """
        return self.pari_mincurve().elltaniyama()

    def cremona_label(self):
        try:
            return self.__cremona_label
        except AttributeError:
            raise errors.UsageError("Cremona label unknown.")

    def torsion_subgroup(self, flag=0):
        """
        Returns the torsion subgroup of this elliptic curve.
        """
        try:
            return self.__torsion
        except AttributeError:
            self.__torsion = self.pari_curve().elltors(flag)
        return self.__torsion

    def newform(self):
        """
        Returns the newform attached to this elliptic curve.
        """
        def comp(n):  # compute series expansion of newform attached E as power series
            v = self.anlist(n)
            return rings.PowerSeries(rings.PowerSeriesRing(rings.Q), v, n)
        M = modform.ModularForms(self.conductor(), 2)
        return modform.Newform(M, comp)

    ## def newform_eval(self, z, prec):
##         """
##         The value of the newform attached to this elliptic curve at
##         the point z in the complex upper half plane, computed using
##         prec terms of the power series expansion.  Note that the power
##         series need not converge well near the real axis.
##         """
##         raise NotImplementedError

    def root_number(self):
        try:
            return self.__root_number
        except AttributeError:
            self.__root_number = int(self.pari_mincurve().ellrootno())
        return self.__root_number

# 	def	weierstrass_p(self):
#         # TODO: add allowing negative valuations for power series
#         return 1/t**2 + a1/t + rings.frac(1,12)*(a1-8*a2) -a3*t \
#                - (a4+a1*a3)*t**2  + O(t**3)

    def padic_sigma(self, p, prec=20):
        """
        Returns the p-adic sigma function of the elliptic curve
        as a power series in t to precision prec.
        """
        R = rings.PowerSeriesRing(variable="t")
        t = R.g0
        a1, a2, a3, a4, a6 = self.__ainvs
        misc.todo("Implement genuine computation of sigma.")
        return t + a1/2 * t**2 + (a1**2+a2)/3 * t**3 + (a1**3+2*a1*a2+2*a3)/4 * t**4 + O(t**5)

    def formal_w(self, prec=20):
        """
        Return the formal power series
        
                     w(t) = t^3 + ...

        to precision O(t^prec) of Proposition IV.1.1 of [Silverman
        AEC1].  This is the formal expansion of w = -1/y about the
        formal parameter t = -x/y at infinity.

        We compute w using the recursive procedure (4.1) on page 18
        of Bluher's `A leisurely introduction to formal groups and
        elliptic curves', which I downloaded from
        
            http://www.math.uiuc.edu/Algebraic-Number-Theory/0076/
            
        """
        try:
            pr, w = self.__formal_w
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = w.parent().gen()
            return w + O(t**prec)
        R = rings.PowerSeriesRing(self.base_field(), "t")
        t = R.gen()
        w = t**3 + O(t**prec)
        a1, a2, a3, a4, a6 = self.ainvs()
        for n in range(4,prec):
            if n<pr:
                w[n] = self.__formal_w[1][n]   # already know it
            else:
                w[n] = a1*w[n-1] + \
                       a2*w[n-2] + \
                       a3*sum([w[i]*w[n-i] for i in range(3,n-2)]) + \
                       a4*sum([w[i]*w[n-1-i] for i in range(3,n-3)]) + \
                       a6*sum([w[i]*w[j]*w[n-i-j] for i in range(3,n-2) \
                               for j in range(3,n-2) if n-i-j >= 3])
        self.__formal_w = (prec, w)
        return w

    def formal_x(self, prec=20):
        """
        Return the formal power series x(t) in terms of the local
        parameter t = -x/y at infinity.
        """
        prec = max(prec,0)
        try:
            pr, x = self.__formal_x
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = x.parent().gen()
            return x + O(t**prec)
        w = self.formal_w(prec+5)
        t = w.parent().gen()
        x = t/w   
        self.__formal_x = (prec, x)
        return x
    
    def formal_y(self, prec=20):
        """
        Return the formal power series y(t) in terms of the local
        parameter t = -x/y at infinity.
        """
        prec = max(prec,0)
        try:
            pr, y = self.__formal_y
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = y.parent().gen()
            return y + O(t**prec)
        w = self.formal_w(prec+6)
        t = w.parent().gen()
        y = -1/w   
        self.__formal_y = (prec, y)
        return y

    def formal_inverse(self, prec=20):
        prec = max(prec,0)
        try:
            pr, inv = self.__formal_inverse
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = inv.parent().gen()
            return inv + O(t**prec)
        x = self.formal_x(prec)
        y = self.formal_y(prec)
        a1, _, a3, _, _ = self.ainvs()
        inv = x / ( y + a1*x + a3)          # page 114 of Silverman, AEC I
        self.__formal_inverse = (prec, inv)
        return inv

    def formal_n_isogeny(self, n, prec=20):
        prec = max(prec,0)
        try:
            pr, phi = self.__formal_n_isogeny[n]
        except AttributeError:
            pr = -1
        if prec <= pr:
            t = phi.parent().gen()
            return phi + O(t**prec)
        try:
            _ = self.__formal_n_isogeny
        except:
            self.__formal_n_isogeny = {}
        misc.todo()

    def formal_group(self, prec=10):
        prec = max(prec,0)
        try:
            pr, F = self.__formal_group
        except AttributeError:
            pr = -1
        if prec <= pr:
            return F
        R1 = rings.PowerSeriesRing(self.base_field(),"t1")
        R2 = rings.PowerSeriesRing(R1,"t2")
        t1 = R1.gen().add_bigoh(prec)
        t2 = R2.gen().add_bigoh(prec)
        w = self.formal_w(prec)
        def tsum(n):
            return sum([t2**m * t1**(n-m-1) for m in range(n)])
        lam = sum([tsum(n)*w[n] for n in range(3,prec)])
        print "lam = ", lam;
        w1 = R1(w.coeffs()).add_bigoh(prec)
        nu = w1 - lam*t1
        a1, a2, a3, a4, a6 = self.__ainvs
        lam2 = lam*lam
        lam3 = lam2*lam
        t3 = -t1 - t2 - \
             (a1*lam + a3*lam2 + a2*nu + 2*a4*lam*nu + 3*a6*lam2*nu)/  \
             (1 + a2*lam + a4*lam2 + a6*lam3)
        inv = self.formal_inverse(prec)
        F = inv(t3)
        self.__formal_group = (prec, F)
        return F

    def formal_mult(self, n, prec=10):
        R = rings.PowerSeriesRing(self.base_field(),"t")
        t = R._0
        if n == 1:
            return t +O(t**prec)
        if n == -1:
            return self.formal_inverse(prec)
        if n < 0:
            return self.formal_inverse(prec)(self.formal_mult(-n,prec))
        F = self.formal_group(prec)
        g = F.parent().base_ring().gen()
        for m in range(2,n+1):
            g = F(g)        
        return R(g.coeffs())

    def discriminant(self):
        # todo: change to use general formula
        # use eval str since could be very large
        return eval(str(self.pari_curve().disc()))

    def j_invariant(self):
        try:
            return self.__j_invariant
        except AttributeError:
            a0 = self.__ainvs[0]
            a1 = self.__ainvs[1]
            a2 = self.__ainvs[2]
            a3 = self.__ainvs[3]
            a4 = self.__ainvs[4]
            self.__j_invariant = (-a0**12 - 24*a0**10*a1 + 72*a0**9*a2 - 240*a0**8*a1**2 + 144*a0**8*a3 + 1152*a0**7*a1*a2 - 1280*a0**6*a1**3 + 2304*a0**6*a1*a3 -1728*a0**6*a2**2 + 6912*a0**5*a1**2*a2 - 6912*a0**5*a2*a3 - 3840*a0**4*a1**4 + 13824*a0**4*a1**2*a3 - 13824*a0**4*a1*a2**2 - 6912*a0**4*a3**2 + 18432*a0**3*a1**3*a2 - 55296*a0**3*a1*a2*a3 + 13824*a0**3*a2**3 - 6144*a0**2*a1**5 + 36864*a0**2*a1**3*a3 - 27648*a0**2*a1**2*a2**2 - 55296*a0**2*a1*a3**2 + 82944*a0**2*a2**2*a3 + 18432*a0*a1**4*a2 - 110592*a0*a1**2*a2*a3 + 165888*a0*a2*a3**2 - 4096*a1**6 + 36864*a1**4*a3 - 110592*a1**2*a3**2 + 110592*a3**3)/(a0**6*a4 - a0**5*a2*a3 + a0**4*a1*a2**2 + 12*a0**4*a1*a4 - a0**4*a3**2 - 8*a0**3*a1*a2*a3 - a0**3*a2**3 - 36*a0**3*a2*a4 + 8*a0**2*a1**2*a2**2 + 48*a0**2*a1**2*a4 - 8*a0**2*a1*a3**2 + 30*a0**2*a2**2*a3- 72*a0**2*a3*a4 - 16*a0*a1**2*a2*a3 - 36*a0*a1*a2**3 - 144*a0*a1*a2*a4 + 96*a0*a2*a3**2 + 16*a1**3*a2**2 + 64*a1**3*a4 - 16*a1**2*a3**2 - 72*a1*a2**2*a3 - 288*a1*a3*a4 + 27*a2**4 + 216*a2**2*a4 + 64*a3**3 + 432*a4**2)
            return self.__j_invariant

    def division_polynomial(self, n):
        misc.todo()

    def formal_log(self, prec=20):
        misc.todo()

    def has_cm(self):
        return self.j_invariant() in [0, 54000, -12288000, 1728, \
                                      287496, -3375, 16581375, 8000, \
                                      -32768,  -884736, -884736000,\
                                      -147197952000, -262537412640768000 ]

    ##########################################################
    # Galois Representations
    ##########################################################
    def is_surjective(self, p, A=1000):
        """
        Return True if the mod-p representation attached to E
        is surjective, False if it is not, or None if we were
        unable to determine whether it is or not.

        INPUT:
            p -- int (a prime number)
            A -- int (a bound on the number of a_p to use)

        OUTPUT:
            a 2-tuple:
            -- surjective or not, or don't know
            -- information about what it might be if not surjective

        EXAMPLES:
        
        """
        if not arith.is_prime(p):
            raise TypeError, "p (=%s) must be prime."%p
        if p == 2:
            invs = self.weierstrass_model().ainvs()
            R = rings.PolynomialRing(self.base_field())
            x = R.gen()
            f = x**3 + invs[3]*x + invs[4]
            if not f.is_irreducible():
                return False, 'rational 2-torsion'
            t, _ = arith.is_square(f.discriminant())
            return not t, 'nonsplit'
        
        if p <= 3:
            return False, None
        
        if self.has_cm():
            return False, "CM"
        an = self.anlist(A)
        ell = 0
        Np = self.conductor() * p
        signs = []
        while True:
            ell = arith.next_prime(ell)
            if ell >= A: break
            if Np % ell != 0:
                a_ell = an[ell]
                if a_ell % p != 0:
                    s = arith.kronecker(a_ell**2 - 4*ell, p)
                    # print ell, s
                    if s == 0: continue
                    if not (s in signs):
                        signs.append(s)
                        if len(signs) == 2:
                            return True, None
            
        # could do something further here...
        return False, signs

    def is_semistable(self):
        if self.base_field() != Q:
            raise NotImplementedError, "is_semistable only implemented for curves over the rational numbers."
        return self.conductor().is_square_free()

    def non_surjective(self, A=1000):
        """
        Returns a list of primes p such that the mod-p representation
        rho_{E,p} *might* not be surjective (this list usually
        contains 2, because of shortcomings of the algorithm).  If p
        is not in the returned list, then rho_{E,p} is provably
        surjective.  If the curve has CM then infinitely many
        representations are not surjective, so we simply return the
        sequence [-1] and do no further computation.

        INPUT:

        OUTPUT:

        EXAMPLES:

        ALGORITHM:
        """
        if self.has_cm():
            return [-1]
        N = self.conductor()
        if self.is_semistable():
            C = 11
        else:
            C = 1 + 4*sqrt(6)*int(N)/3 * mul([1+1.0/int(p) for p,_ in factor(N)])
        B = []
        p = 2
        while p <= C:
            t, v = self.is_surjective(p, A=A)
            if not t:
                B.append((p,v))
            p = next_prime(p)
        return B

    def xxx_eval_modular_form(self, z, prec):
        an = self.anlist(prec)
        s = 0
        z = C(z)
        I = C.gen()
        pi = transcendental.pi()
        r0 = (2*pi*I*z).exp()
        r = r0
        for n in xrange(prec):
            s += an[n]*r      
            r *= r0
        return s

    def xxx2_eval_modular_form(self, z, prec):
        an = self.anlist(prec)
        s = 0
        z = pari(z)
        I = pari("I")
        pi = pari("Pi")
        r0 = (pari(2)*pi*I*z).exp()
        r = r0
        for n in xrange(1,prec):
            s += an[n]*r      
            r *= r0
        return s.python()

    def eval_modular_form(self, points, prec):
        if not isinstance(points, (list,xrange)):
            try:
                points = list(points)
            except TypeError:
                return self.eval_modular_form([points],prec)
        an = self.pari_mincurve().ellan(prec)
        s = 0
        I = pari("I")
        pi = pari("Pi")
        c = pari(2)*pi*I
        ans = []
        for z in points:
            s = pari(0)
            r0 = (c*z).exp()
            r = r0
            for n in xrange(1,prec):
                s += an[n-1]*r      
                r *= r0
            ans.append(s.python())
        return ans

    def weierstrass_model(self):
        a0, a1, a2, a3, a4 = self.__ainvs
        return EllipticCurve([-27*a0**4 - 216*a0**2*a1 + 648*a0*a2 - 432*a1**2 + 1296*a3, \
                              54*a0**6 + 648*a0**4*a1 - 1944*a0**3*a2 + 2592*a0**2*a1**2 - 3888*a0**2*a3 - 7776*a0*a1*a2 + 3456*a1**3 - 15552*a1*a3 + 11664*a2**2 + 46656*a4])
    
class Point:

    def __init__(self, curve, x=oo, y=oo, check=True):
        if not isinstance(curve, EllipticCurve):
            raise errors.TypeError("Argument 1 must be an elliptic curve.")
        self.__curve = curve
        self.is_vector = True
        K = curve.base_field()
        if isinstance(x, (list,tuple)):
            if len(x) == 1:
                x, y = x[0], 1
            elif len(x) == 2:
                x, y = x[0], x[1]
            else:
                if x[2] == 0:
                    x = rings.infinity
                else:
                    z = K(x[2])
                    x, y = K(x[0])/z, K(x[1])/z
        if x is oo:
            self.__x, self.__y = oo, oo
            return
        x = K(x)
        y = K(y)
        self.__x, self.__y = x, y
        if y*y + curve.a1()*x*y + curve.a3()*y != \
               x**3 + curve.a2()*x**2 + curve.a4()*x + curve.a6():
            raise errors.MathError("Point %s is not on curve."%self)

    def __repr__(self):
        if self.is_infinity():
            return "(1,0,0)"
##         if isinstance(self.base_field(), rings.RationalField):
##             d = self.__x.denominator().lcm(self.__y.denominator())
##         else:
##             d = self.base_field()(1)
        return "(%s, %s)"%(self.__x, self.__y)

    def base_field(self):
        return self.__curve.base_field()

    def order(self):
        """
        Return the order of this point on the elliptic curve.
        If the point has infinite order, returns 0.

        EXAMPLE:
            >>> E = EllipticCurve([0,0,1,-1,0])
            >>> P = E([0,0]); P
            (0, 0)
            >>> P.order()
            0

            >>> E = EllipticCurve([0,1])
            >>> P=E([-1,0])
            >>> P.order()
            2
        """
        if self.is_infinity():
            return 1
        return self.__curve.pari_curve().ellorder([self.__x, self.__y])

    def xy(self):
        return self.__x, self.__y

    def x(self):
        return self.__x

    def y(self):
        return self.__y

    def curve(self):
        return self.__curve

    def __getitem__(self, n):
        if n==0:
            return self.__x
        elif n==1:
            return self.__y
        else:
            raise IndexError, "Index must be 0 or 1"

    def is_infinity(self):
        return self.__x is rings.infinity
    
    def __add__(self, right):
        """Add self to right."""
        if self.curve() != right.curve():
            raise errors.MathError("Points must be on the same curve.")

        # Use Prop 7.1.7 of Cohen "A Course in Computational Algebraic Number Theory"
        if self.is_infinity():
            return right
        if right.is_infinity():
            return self
        a1, a2, a3, a4, a6 = self.__curve.ainvs()
        x1, y1 = self.__x, self.__y
        x2, y2 = right.__x, right.__y
        if x1 == x2 and y1 == -y2 - a1*x2 - a3:
            return Point(self.__curve) # point at infinity
        if x1 == x2 and y1 == y2:
            m = (3*x1*x1 + 2*a2*x1 + a4 - a1*y1) / (2*y1 + a1*x1 + a3)
        else:
            m = (y1-y2)/(x1-x2)
        x3 = -x1 - x2 - a2 + m*(m+a1)
        y3 = -y1 - a3 - a1*x3 + m*(x1-x3)
        return Point(self.__curve, x3, y3)

    def __sub__(self, right):
        return self + (-right)

    def __neg__(self):
        if self.is_infinity():
            return self
        E, x, y = self.__curve, self.__x, self.__y
        return Point(self.__curve, x, -y - E.a1()*x - E.a3())

    def __mul__(self, m):
        return self.__rmul__(m)

    def __rmul__(self, m):
        m = int(m)
        if m<0:
            return (-m)*(-self)
        if m==1:
            return self
        if m==0:
            return Point(self.__curve)
        power = Point(self.__curve)
        i = 0
        apow2 = self
        while ((m>>i) > 0):
            if((m>>i) & 1):
                power += apow2
            apow2 += apow2
            i += 1
        return power

    def sigma(self, p):
        k = rings.Qp(p)
        if self.is_infinity():
            return k(0)
        sigma = self.__curve.sigma(p)
        return sigma(k(-self.__x/self.__y))
        

    def padic_height(self, p):
        """Returns the p-adic cyclotomic height of the point.

        padic_height(p)

        Input:
           p: a prime number
        Output:
           p-adic cyclotomic height, which is a single p-adic number.

        Algorithm:
        We compute this height using the following formula, which
        is valid for points that are in the intersection of the
        identity component of the Neron model with the kernel of
        reduction modulo p:

     h(P) = 1/2 * sum_{ell!=p} sup(0,-ord_ell(x(P)))  + log_p(sigma_p(-x(P)/y(P)) / e),

        where P=(a/e^2, b/e^3) with gcd(a,e)=1, and
        where the first sum is over primes ell that don't equal p.
        If P isn't in the subgroup mentioned above, let n be a positive
        integer so that n*P is in that subgroup.  Then we return
        h(n*P)/(n**2), which does not depend on the choice of n, and
        is defined using the above formula.

        """

        if not arith.is_prime(p):
            raise errors.UsageError("p must be a prime number.")
        E = self.curve()
        if E.conductor() % p == 0:
            raise errors.UsageError("p must be a prime of good reduction.")

        ap = E.ap(p)
        if ap == 0:
            raise errors.UsageError("p must a prime of good ordinary reduction.")
        
        Np = p+1 - E.ap(p)
        c = 1
        for q, _ in arith.factor(E.conductor()):
            c = arith.lcm(c, E.tamagawa_number(q))

        n = Np*c
        misc.verbose("Multiplying point by %s."%n)
        R = n*self
            
        x, y = R.xy()
        sigma = E.padic_sigma(p)
        Qp = rings.pAdicField(p)
        print "x = ", x
        print "-x/y = ", Qp(-x/y)
        d = x.denominator()
        e = d.sqrt()
        if e*e != d:
            raise errors.MathError("The denominator is not a perfect square!?")
        t = Qp(-x/y)
        return (sigma(t)/Qp(e)).log()



        w = Qp(0)
        for ell, ord in d.factor():
            if ell != p:
                w += ord * Qp(ell).log()
        # note that there is no 1/2, since we factored the square root.
        t = Qp(-x/y)
        print "sigm = ", sigma(t)
        print "e = ", e
        hP = w + (sigma(t)/e).log()
        print "w = ", w
        print "sigma(t) = ", sigma(t)
        print "sigma(t)/e = ", sigma(t)/e
        print "(sigma(t)/e).log() = ", (sigma(t)/e).log()
        print "sum = ", w + (sigma(t)/e).log()

        return hP/Qp(n**2)


def curves(conductors):
    """
    Return all known curves (in database) with conductor in the list
    of conductors, grouped by isogeny class.
    """
    v = sage.tables.elliptic_curves.Database().curves(conductors)
    return [EllipticCurve(E) for E in v]

def optimal_curves(conductors):
    """
    Return all known optimal curves (in database) with conductor in the list
    of conductors.
    """
    v = sage.tables.elliptic_curves.Database().optimal_curves(conductors)
    return [EllipticCurve(E) for E in v]


                
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
